@@ -8,6 +8,9 @@ svychisq<-function(formula, design,
   if (ncol(attr(terms(formula),"factors"))>2)
     stop("Only 2-way tables at the moment")
   statistic<-match.arg(statistic)
+
+  if(!is.null(design$postStrata))
+    warning("Post-stratification not implemented")
   
   rows<-formula[[2]][[2]]
   cols<-formula[[2]][[3]]
